/*bsplines-hack.c - John Cantrell - jdcantrell@gmail.com*/
/* bsplines, Copyright (c) 2003 John Cantrell*/
/* dangerball, Copyright (c) 2001, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*NOTES:
 * Parts of this program were borrowed from dangerball, gears and atunnel. 
 * The bspline algorithm is one based off the one given in the class
 * EECE636 - Intro to Computer Graphics by Kenneth Carepenter at Kansas
 * State University
 */

/*REVISION HISTORY:
 * 05/04/03 - Initial Revision
 * 05/06/03 - Made the curves plot in 3d
 * 06/09/03 - Initial attempt at converting into xscreensaver hack
 * 06/14/03 - Added more options 
 * 06/29/03 - made depth testing optional
 * 08/18/04 - ported over texture code and updates from current standalone
 *
 * todo
 * mess with color blending
 * find ways to speed it up
 * 
*/

/*************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <X11/Intrinsic.h>
/*************************************************************/
extern XtAppContext app;

#define PROGCLASS	"B-Splines"
#define HACK_INIT	init
#define HACK_DRAW	display
#define HACK_RESHAPE	reshape
#define sws_opts	xlockmore_opts

#define DEF_STEPS       "30"
#define DEF_CURVES      "1"
#define DEF_SPEED       "10"
#define DEF_POINTS      "10"
#define DEF_ANTIALIAS 	"True"
#define DEF_LINE		"True"
#define DEF_SIZE        "2"
#define DEF_DOTEXTURE "False"


#define DEFAULTS "*delay:			30000       \n" \
			"*showFPS:      	false      	\n" \
			"*antialias:" 		DEF_ANTIALIAS	"\n" \
			"*speed:    " 		DEF_SPEED 	" \n" \
			"*steps:    " 		DEF_STEPS   "\n" \
			"*curves:   " 		DEF_CURVES 	"\n" \
			"*controlPoints:" 	DEF_POINTS 	"\n" \
			"*lines:" 			DEF_LINE 	"\n" \
			"*size:" 			DEF_SIZE 	"\n" \
            "*dotexture:" 			DEF_DOTEXTURE 	"\n" 
#define sws_opts	xlockmore_opts
#undef countof
#define countof(x) (sizeof((x))/sizeof((*x)))
#define randomNumber while ((randomNum = (random()/(float)RAND_MAX))<=.0001) {}		

#include "xlockmore.h"
#include "gltrackball.h"
#include "xpm-ximage.h"
#include "../images/particle.xpm"
#define I_HAVE_XPM
#include <ctype.h>
#include <GL/glu.h>
	
static int speed;
static int curves;
static int steps;
static int controlPoints;
static float size;
static Bool antialias;
static Bool lines;
static Bool dotexture;

static XrmOptionDescRec opts[] = {
  { "-speed",  ".speed",  XrmoptionSepArg, 0 },
  { "-curves", ".curves", XrmoptionSepArg, 0 },
  { "-smoothness", ".smoothness", XrmoptionSepArg, 0 },
  { "-curviness", ".curviness", XrmoptionSepArg, 0 },
  { "-noantialias", ".noantialias", XrmoptionNoArg, "false" },
  { "-points", ".points", XrmoptionNoArg, "false" },
  {"-dotexture",".dotexture",XrmoptionNoArg,"true"},
  { "-size", ".size", XrmoptionSepArg, 0 }
  
};

static argtype vars[] = {
  {&curves, "curves", "Curves", DEF_CURVES, t_Int},
  {&speed,     "speed",  "Speed",  DEF_SPEED,  t_Int},
  {&steps,     "smoothness",  "Smoothness",  DEF_STEPS,  t_Int},
  {&controlPoints,     "curviness",  "Curviness",  DEF_STEPS,  t_Int},
  {&antialias,     "noantialias",  "No Antialias",  DEF_ANTIALIAS,  t_Bool},
  {&lines,     "points",  "Points",  DEF_LINE,  t_Bool},
  {&size,     "size",  "Size",  DEF_SIZE,  t_Float},
  {&dotexture,     "dotexture", "Do Texture", DEF_DOTEXTURE, t_Bool},
};

ModeSpecOpt sws_opts =  {countof(opts), opts, countof(vars), vars, NULL};

typedef struct {
  GLXContext *glx_context;
  int ncolors;
  XColor *colors;
  int ccolor;
  int color_shift;
} screen_stuff;

static screen_stuff *screen = NULL;

struct lineList {
	float *x,*y,*z,*dx,*dy,*dz;
	float r1,g1,b1;
	float rinc,binc,ginc;
	int numberOfPoints;
	struct lineList* previous;
	struct lineList* next;	
};

int width, height;
struct lineList* head = NULL;
struct lineList* tail = NULL;

/*blending poly tables*/
float* poly1;
float* poly2;
float* poly3;
float* poly4;
float texOffset;
GLuint texture;
/******************************************************************************************/
static void LoadTexture(ModeInfo * mi, char **fn)
{
#if defined( I_HAVE_XPM )
	XImage *teximage;    /* Texture data */
 
        if ((teximage = xpm_to_ximage(MI_DISPLAY(mi), MI_VISUAL(mi),
			 MI_COLORMAP(mi), fn)) == None) {
	    (void) fprintf(stdout, "Error reading the texture.\n");
	    glDeleteTextures(1, &texture);
            dotexture = False;
#ifdef STANDALONE
	    exit(0);
#else
	    return;
#endif
	}

#ifdef HAVE_GLBINDTEXTURE
	glBindTexture(GL_TEXTURE_2D, texture);
#endif /* HAVE_GLBINDTEXTURE */
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	clear_gl_error();
    
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, teximage->width, teximage->height, 
			0, GL_RGBA, GL_UNSIGNED_BYTE, teximage->data);
	check_gl_error("texture");

	/* Texture parameters, LINEAR scaling for better texture quality */
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 

	XDestroyImage(teximage);
#else /* !I_HAVE_XPM */
	dotexture = False;
#endif /* !I_HAVE_XPM */
}

/*************************************************************/	
/* adds a new b-spline loop 
 * first 6 parameters are rgb values for colors
 * last parameter is the number of control points for b-spline drawing
*/
void addLoop(
	float nr1,float ng1,float nb1,float nr2,float ng2,float nb2,int points) {
	int t=0,a=0;
	float temp[points+5];
	struct lineList* new;
	new = malloc(sizeof(struct lineList));
    
	/*we add 5 extra points to redraw the fist 4 to get a nice smooth loop*/
	float *newx = malloc(sizeof(temp)); 
	float *newy = malloc(sizeof(temp));
	float *newz = malloc(sizeof(temp));
	float *newdx = malloc(sizeof(temp));
	float *newdy = malloc(sizeof(temp));
	float *newdz = malloc(sizeof(temp));
	
	/* add an array of new points */
	for (t=0;t<points;t++) {
		float randomNum=0;
		randomNumber/* macro defined above*/
		*(newx+t)=randomNum*2-1;
		randomNumber
		*(newy+t)=randomNum*2-1;
		randomNumber
		*(newz+t)=randomNum*1;
		randomNumber
		*(newdx+t)= (randomNum*2-1)/(10*speed);
		randomNumber
		*(newdy+t)= (randomNum*2-1)/(10*speed);
		randomNumber
		*(newdz+t)= (randomNum*2-1)/(10*speed);
	}
	/*add the last few points to make a loop */
	
	for (t=t;t<points+5;t++) {
		*(newx+t)=*(newx+a);
		*(newy+t)=*(newy+a);
		*(newz+t)=*(newz+a);
		*(newdx+t)= *(newdx+a);
		*(newdy+t)= *(newdy+a);
		*(newdz+t)= *(newdz+a);
		a++;
	}
	
	/*add to the linelist struct*/
	new->x=newx;
	new->y=newy;
	new->z=newz;
	new->dx=newdx;
	new->dy=newdy;
	new->dz=newdz;
	
	/* set colors */
	new->r1=nr1;
	new->g1=ng1;
	new->b1=nb1;

	/*calculate color blending increments */
	new->rinc = (nr2-nr1)/(.5*(points)*steps);
	new->ginc = (ng2-ng1)/(.5*(points)*steps);
	new->binc = (nb2-nb1)/(.5*(points)*steps);
	
	new->numberOfPoints=points+4;

	/*set links */
	new->previous=tail;
	new->next=NULL;
	if (head==NULL) {head=new;}
	if (tail!=NULL) {tail->next=new;}
	tail=new;	
}
/*************************************************************/
void display(ModeInfo *mi) {

  	Display *dpy = MI_DISPLAY(mi);
  	Window window = MI_WINDOW(mi);
	float r, g, b, py, px,pz,rinc,binc,ginc,px1=0.,py1=0.,pz1=0.;
	float  *x, *y, *dx, *dy,*z,*dz;
	int numPoint=0,points;
	struct lineList* currentLine;
    int t,s;			
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/*cycle through each curve*/
	for(currentLine=head;currentLine!=NULL;currentLine=currentLine->next) {
		numPoint=0;/*current point number*/
		px1=-1;/* first point unset*/
		
		x=currentLine->x;
		y=currentLine->y;
		z=currentLine->z;
		dx=currentLine->dx;
		dy=currentLine->dy;
		dz=currentLine->dz;
		rinc = currentLine->rinc;
		ginc = currentLine->ginc;
		binc = currentLine->binc;
		r=currentLine->r1;
		g=currentLine->g1;
		b=currentLine->b1;
		points = currentLine->numberOfPoints;

		/*begin drawing curve*/
		
		if (!dotexture) {
            if (lines)	{glBegin(GL_LINE_STRIP);}  
            else {glBegin(GL_POINTS);}
        }
		for (t=0;t<points-4;t++) {
			for (s=0;s<steps;s++) {
			/*find x,y,z pointer math for no reason*/
		px = ((*x)*(*(poly1+s)) + *(x+1)*(*(poly2+s)) + *(x+2)*(*(poly3+s))
				+ *(x+3)*(*(poly4+s)));
		py = ((*y)*(*(poly1+s)) + *(y+1)*(*(poly2+s)) +	*(y+2)*(*(poly3+s)) 
				+ *(y+3)*(*(poly4+s)));		
		pz = ((*z)*(*(poly1+s)) + *(z+1)*(*(poly2+s)) + *(z+2)*(*(poly3+s)) 
				+ *(z+3)*(*(poly4+s)));
				/*inc colors*/
				if (((float)numPoint/points)<.3) {r+=rinc;b+=binc;g+=ginc;}
				else {r-=rinc;g-=ginc;b-=binc;}
				/*glColor3f(r,g,b);*/
                glColor4f(r,g,b,(pz+1)/2);
				if (px1==-1) {px1=px;py1=py;pz1=pz;}
				
				if (dotexture) {
                    glBegin(GL_TRIANGLE_STRIP);		
				glTexCoord2d(1, 1);
                glVertex3f(px + texOffset, py + texOffset, pz);
                glTexCoord2d(0, 1);
                glVertex3f(px - texOffset, py + texOffset, pz);
                glTexCoord2d(1, 0);
                glVertex3f(px + texOffset, py - texOffset, pz);
                glTexCoord2d(0, 0);
                glVertex3f(px - texOffset, py - texOffset, pz);
    	        glEnd();
                }
                else 
                {
                    glVertex3f(px,py,pz);
                }
				
			
			}
			
			/*update points */
			*x+=*dx,*y+=*dy,*z+=*dz;
			if (*x>1||*x<-1) *dx=-*dx;
			if (*y>1||*y<-1) *dy=-*dy;
			if (*z>1||*z<-1) *dz=-*dz;
			numPoint++,x++,dx++,y++,z++,dz++,dy++;
		}
		if (!dotexture) {
		    glVertex3f(px1,py1,pz1);
		    glEnd();
        }
		
		/*update the last few points */		
		for (t=0; t<4;t++) {
			*x+=*dx,*y+=*dy,*z+=*dz;
			if (*x>1||*x<-1) *dx=-*dx;
			if (*y>1||*y<-1) *dy=-*dy;
			if (*z>1||*z<-1) *dz=-*dz;
			x++,dx++,y++,dy++,z++,dz++;
		}
	}
	if (mi->fps_p) do_fps (mi);
	glFinish();
	glXSwapBuffers(dpy, window);
}
/*************************************************************/
void reshape(ModeInfo *mi, int w, int h) {
	width=w;
	height=h;
	glLoadIdentity();
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glFrustum(-1.0,1.0,-1.0,1.0,-15.0,15.0);
	glMatrixMode(GL_MODELVIEW);
}
/*************************************************************/
void init (ModeInfo *mi) {
	float s;int p=0;
	screen_stuff *sp;
  	
	if (!screen) {
  	  	screen = (screen_stuff *)
      	calloc (MI_NUM_SCREENS(mi), sizeof (screen_stuff));
		if (!screen) {
      		fprintf(stderr, "%s: out of memory\n", progname);
      		exit(1);
    	}
      	sp = &screen[MI_SCREEN(mi)];
  	}

	sp = &screen[MI_SCREEN(mi)];
	sp->glx_context = init_GL(mi);

	if (speed<=0) speed=1;
			
	reshape (mi, MI_WIDTH(mi), MI_HEIGHT(mi));
	
	glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);
	glFrustum(-1.0,1.0,-1.0,1.0,-5.0,10.0);
	glMatrixMode(GL_MODELVIEW);
	glClearColor (0.0, 0.0, 0.0, 0.0);
    
	/*enable smoothing */
	if (antialias) {
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH); 
	glEnable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);			
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
		}
	
	/*set sizes*/	
	glLineWidth(size);
	glPointSize(size);
		
	/*init blending poly tables*/
	
	poly1 = malloc(sizeof(float)*(steps+5));
	poly2 = malloc(sizeof(float)*(steps+5));
	poly3 = malloc(sizeof(float)*(steps+5));
	poly4 = malloc(sizeof(float)*(steps+5));
	/*these are our blending polynomials. they make a neat graph where at any
	 *given point all four add up to one */
	for (s=0;s<=1+3*(1.0/steps);s+=1.0/steps) {
		*(poly1+p)=(1.0/6)*((1-s)*(1-s)*(1-s));
		*(poly2+p)=(1.0/6)*((s+2)*(1-s)*(1-s)+(s+1)*(1-s)*(2-s)+s*(2-s)*(2-s));
		*(poly3+p)=(1.0/6)*((s+1)*(s+1)*(1-s)+s*(s+1)*(2-s)+s*s*(3-s));
		*(poly4+p)=(1.0/6)*s*s*s;
		p++;
	}
	if (dotexture) {
		texOffset = 13.0/width;
        glGenTextures(1, &texture);
        LoadTexture(mi, particle_xpm);
    }
        /*add lines - dirty but it works*/
	if (controlPoints<2) controlPoints=2;
	addLoop(0.0,1.0,1.0,0.0,0.0,1.0,controlPoints);
	if (curves>=2) {addLoop(1.0,1.0,0.0,1.0,0.0,0.0,controlPoints);}
	if (curves>=3) {addLoop(1.0,1.0,1.0,0.0,1.0,0.0,controlPoints);}
	if (curves>=4) {addLoop(1.0,1.0,1.0,0.45,0.0,0.45,controlPoints);}
}
